;; Generated package description from hcel.el  -*- no-byte-compile: t -*-
(define-package "hcel" "1.0.0.0.20221006.55158" "Haskell codebase explorer / cross referencer" '((emacs "28")) :commit "f1d30dc711f50b988a8305029b43d13ff0cb5321" :authors '(("Yuchen Pei" . "id@ypei.org")) :maintainer '("Yuchen Pei" . "id@ypei.org") :keywords '("haskell") :url "https://g.ypei.me/hc.el.git")
